! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
      subroutine mati(arr,norder,det)

      use param
      use dal
      use emldata

  !----------------------------------------------------------
  ! Matrix inversion.
  ! Based on subroutine mati used by XMMSAS emldetect.
  ! Adapted to Fortran 90 (replaced labels, arithmetic-if, 
  ! and goto statements. (GL 12/09/2011)
  ! Replaces norder*norder elements of matrix arr by
  ! inverse matrix. 
  !----------------------------------------------------------
  

    integer                :: k,norder,i,j,l
    real(kind=double)      :: arr(max_par,max_par)
    real(kind=double)      :: amax,save,det
    integer                :: ik(max_par),jk(max_par),imax,jmax
    logical                :: matzero

!    real(kind=real32)                  :: t1,t2
!    integer(kind=int32)                :: icount
!    real(kind=real64), dimension(:,:),allocatable  :: arrsave

    det=1.D0


    matzero=.false.



    kloop: do k=1,norder


    ! find largest element arr(i,j) in rest of matrix.   
      amax=0.D0
      jmaxloop: do 
        imaxloop: do
          do i=k,norder
            do j=k,norder
              if (abs(arr(i,j)) >= abs(amax)) then
                amax=arr(i,j)
                ik(k)=i
                jk(k)=j
              endif
            enddo
          enddo
          if(amax == 0.D0) then
            det = 0.D0
            matzero = .true.
            exit kloop
          endif


         ! interchange rows and columns to put amax in arr(k,k)
          imax=ik(k)
          if (imax >= k) exit imaxloop
        end do imaxloop 
        if(imax > k) then
          do j=1,norder
            save=arr(k,j)
            arr(k,j)=arr(imax,j)
            arr(imax,j)=-save
          enddo
        endif
        jmax=jk(k)
        if (jmax >= k) exit jmaxloop
      end do jmaxloop
      if(jmax > k) then
        do i=1,norder
          save=arr(i,k)
          arr(i,k)=arr(i,jmax)
          arr(i,jmax)=-save
        enddo
      endif

      ! accumulate elements of inverse matrix
      do i=1,norder
        if (i /= k) arr(i,k) = -arr(i,k)/amax
      enddo
      do i=1,norder
        do j=1,norder
          if (i /=k ) then
            if(j /= k) then
              arr(i,j)=arr(i,j)+arr(i,k)*arr(k,j)
            endif
          endif
        enddo
      enddo
      do j=1,norder
        if(j /= k) arr(k,j)=arr(k,j)/amax
      enddo
      arr(k,k)=1./amax
      det=det*amax
    end do kloop

    if (.not. matzero) then
      ! restore ordering of matrix
      do l=1,norder
        k=norder-l+1
        j=ik(k)
        if (j > k) then
          do i=1,norder
            save=arr(i,k)
            arr(i,k)=-arr(i,j)
            arr(i,j)=save
          enddo     
        endif     
        i=jk(k)
        if (i > k) then
          do j=1,norder
            save=arr(k,j)
            arr(k,j)=-arr(i,j)
            arr(i,j)=save
          enddo
        endif
      enddo
    endif


      end subroutine mati



